from typing import Callable, Generic, Hashable, Tuple, TypeVar


T = TypeVar('T')


class DistinctFilter(Generic[T]):
    def __init__(self, f:Callable[[T], Hashable]) -> None:
        self.__encounters = set()
        self.__f = f


    def __call__(self, val: T) -> bool:
        __val = self.__f(val)
        if __val in self.__encounters:
            return False
        self.__encounters.add(__val)
        return True


class Rangifier(Generic[T]):
    def __init__(self, init_val: T) -> None:
        self.__prev: T = init_val

    def __call__(self, val: T) -> Tuple[T, T]:
        res = (self.__prev, val)
        self.__prev = val
        return res


from typing import Any, Callable, Dict, Hashable, Iterable, Optional, Sequence, TypeVar

from trent.nth import __no_value, nth




# ======================================================

_T = TypeVar('_T')
_T2 = TypeVar('_T2')

NV = __no_value()

def identity(val: _T) -> _T:
    return val


# ============================================================================
#               UTIL

def __get(m: Dict[Any, _T], key, default: _T2 = None) -> _T|_T2:
    return m.get(key, default)
    

def getter(key, default: _T2 = None) -> Callable[[Dict[Any, _T]], _T|_T2]:
    def __f(__m: Dict[Any, _T]) -> _T | _T2:
        return __get(__m, key, default)
    return __f


def gtr(key, default: _T2 = None, /) -> Callable[[Dict[Any, _T]], _T|_T2]:
    return getter(key, default)

def gtr_(key:Hashable, _type: type[_T], /) -> Callable[[Dict[Any, Any]], _T]:
    def __f(__m: Dict[Any, _T]):
        _res = __get(__m, key, NV)
        if isinstance(_res, __no_value):
            raise Exception(f'Missing value by key: {key}, from map: {__m}')
        return _res
    return __f


def isnone(val: Any) -> bool:
    return val is None


if __name__ == '__main__':
    d = {'foo': 1}
    
    f = gtr_('foo', int)
    
    res = f(d)


from .coll import icoll
from .nth import first, first_, second, second_
from .func import gtr, gtr_
from .interface import (
    cat,
    catmap,
    cfilter,
    cmap,
    seq,
    coll,
    groupcoll,
    groupmap,
    map_to_pair,
    mapcat,
    pairmap,
    pmap,
    pmap_,
    rangify,
)

__all__ = [
    'icoll',
    
    'seq',
    'coll',
    'cmap', 
    'cfilter',
    'pmap',
    'pmap_',
    'cat',
    'mapcat',
    'catmap',
    'pairmap',
    'groupmap',
    'groupcoll',
    'map_to_pair',
    'rangify',
    
    'first',
    'first_',
    'second',
    'second_',
    
    'gtr',
    'gtr_'
]


from typing import Any, Callable, Dict, Iterable, Optional, Tuple, TypeVar, overload

from trent.coll import icoll
from trent.concur import CPU_COUNT


T = TypeVar('T')
T1 = TypeVar('T1')
T2 = TypeVar('T2')


@overload
def seq() -> icoll[Any]:...
@overload
def seq(seq: Dict[T1, T2]) -> icoll[Tuple[T1, T2]]: ...
@overload
def seq(seq: Iterable[Tuple[T1, T2]]) -> icoll[Tuple[T1, T2]]: ...
@overload
def seq(seq: Iterable[T]) -> icoll[T]: ...

def seq(seq: Optional[Iterable] = None) -> icoll:
    if isinstance(seq, Dict):
        return icoll(seq.items())
    return icoll(seq)

coll = seq


def cmap( f: Callable[[T], T2], seq: Optional[Iterable[T]]) -> icoll[T2]:
    return icoll(seq).map(f)


def cfilter(pred: Callable[[T], Any], seq: Optional[Iterable[T]]) -> icoll[T]:
    return icoll(seq).filter(pred)


def pmap(f: Callable[[T], T2], seq: Optional[Iterable[T]]) -> icoll[T2]:
    return icoll(seq).pmap(f)


def pmap_(f: Callable[[T], T2], seq: Optional[Iterable[T]], threads: int = CPU_COUNT) -> icoll[T2]:
    return icoll(seq).pmap_(f, threads)


def cat(seq: Optional[Iterable[Iterable[T]]]) -> icoll[T]:
    return icoll(seq).cat()


def mapcat(f: Callable[[T], Iterable[T2]], seq: Optional[Iterable[T]]) -> icoll[T2]:
    return icoll(seq).mapcat(f)


def catmap(f: Callable[[Any], T2], seq: Optional[Iterable[Iterable[T]]]) -> icoll[T2]:
    return icoll(seq).catmap(f)


def pairmap(f:Callable[[T1, T2], T], seq: Iterable[Tuple[T1, T2]], ) -> icoll[T]:
    return icoll(seq).pairmap(f)


def groupcoll(seq:Iterable[Tuple[T1, Iterable[T2]]]) -> icoll[Tuple[T1, T2]]:
    return icoll(seq).groupmap()


def groupmap(f:Callable[[T1, T2], T], seq:Iterable[Tuple[T1, Iterable[T2]]]) -> icoll[T] | icoll[Tuple[T1, T2]]:
    return icoll(seq).groupmap(f)


@overload
def map_to_pair(seq: Iterable[T], f_key:Callable[[T], T1]) -> icoll[Tuple[T1, T]]: ...
@overload
def map_to_pair(seq: Iterable[T], f_key:Callable[[T], T1], f_val: Callable[[T], T2]) -> icoll[Tuple[T1, T2]]: ...

def map_to_pair(seq: Iterable[T], f_key:Callable[[T], T1], f_val: Optional[Callable[[T], T2]] = None) -> icoll[Tuple[T1, T2]] | icoll[Tuple[T1, T]]:
    if f_val is not None:
        return icoll(seq).map_to_pair(f_key, f_val)
    return icoll(seq).map_to_pair(f_key)



def rangify(seq: Iterable[T]) -> icoll[Tuple[T, T]]:
    return icoll(seq).rangify()


if __name__ == '__main__':
    pass